#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include "MNLClient.h"

int gSkt = 0;
struct sockaddr_in gDestSockAddr;

/** Network byte order */
uint16_t gMyUDPPortN = 0;

/** The destination port - in host byte order */
uint16_t gDestPortH = 0;

void InitSkt() {

  struct sockaddr_in lMyAddr;
  socklen_t lMyAddrLen = sizeof(lMyAddr);
  
  gSkt = socket (PF_INET, SOCK_DGRAM, 0);
  if (gSkt < 0) {
    printf (
        "Error: Unable to open socket : %s\n",
        strerror(errno)
      );
    exit(1);
  }
 

  // Set the destination to which we are sending packets.
  gDestSockAddr.sin_family = AF_INET;    
  gDestSockAddr.sin_port = htons(gDestPortH);
  gDestSockAddr.sin_addr.s_addr = INADDR_ANY; 
  memset(&(gDestSockAddr.sin_zero), '\0', sizeof(gDestSockAddr.sin_zero));

  // Bind to a random port
  lMyAddr.sin_family = AF_INET; 
  lMyAddr.sin_port = 0;
  lMyAddr.sin_addr.s_addr = INADDR_ANY;
  memset(&(lMyAddr.sin_zero), '\0', sizeof(lMyAddr.sin_zero));

  if (bind( gSkt , 
            (struct sockaddr *)&lMyAddr, 
            lMyAddrLen
            ) == -1) {

    printf ( "Error: Unable to bind udp listen socket to a random port: %s\n",
              strerror (errno)
            );
    exit(1);
  }

  // Get the port on which we are listening.
  if ( getsockname (
                    gSkt,
                    (struct sockaddr *) &lMyAddr, 
                    &lMyAddrLen
                    ) != -1) {

    printf("My socket is bound to %s:%d\n",
                    inet_ntoa (lMyAddr.sin_addr),
                    ntohs (lMyAddr.sin_port)
                );
    gMyUDPPortN = lMyAddr.sin_port;

  } else {
    printf ("Error: Unable to get port number of my socket :%s\n",
            strerror(errno)
          );
    exit(1);
  }



}

void WaitForReply() {
  char recvBuf [10000];
  int recvBufLen = sizeof(recvBuf);
  
  struct sockaddr_in remoteAddr;
  socklen_t addrLen = sizeof(struct sockaddr_in);
  int numBytesRead = recvfrom (gSkt, recvBuf, recvBufLen,0, (struct sockaddr *)&remoteAddr, &addrLen);
  if (numBytesRead == -1) {
      printf ("Error while receiving UDP packet : %s.\n", strerror (errno));
      return;
  }
      
  if (numBytesRead > 0) {
      // Some bytes were actually read.
      // Echo it back
      
      // sanitize the received string in order to prevent buffer overflow if
      // dangerous string functions like strcpy are used.
      recvBuf [recvBufLen - 1] = '\0';
      printf ("REPLY RECD: %s:%d Received %d bytes over UDP: %s\n", 
              inet_ntoa (remoteAddr.sin_addr),
              ntohs(remoteAddr.sin_port),
              numBytesRead,
              recvBuf
            );
  } else {
      printf ("REPLY RECD: %s:%d Warning: 0 bytes received.\n", 
                          inet_ntoa (remoteAddr.sin_addr),
                          ntohs (remoteAddr.sin_port)
                  );
  }

}

void SendPkts(int iPktSize, int iMaxCount, int iWaitForResponse) {
  int lPktCount = 0;
  int lPktCountMax = iMaxCount;

  int lNumBytesSent = 0;
  int lErrFlag = 0;

  char lpPkt[MAX_PKT_LEN];
  int lPktSize = 0;

  

  for (lPktCount = 0; lPktCount < lPktCountMax; lPktCount++) {
 
    if (iPktSize > 0) {
      char lA = 'c';
      int lPktFillIndex;
      for (lPktFillIndex = 0; lPktFillIndex < iPktSize-1; lPktFillIndex++) {
        lpPkt[lPktFillIndex] = lA + (lPktCount % 10);
      }
      lpPkt[lPktFillIndex] = '\0';
      lPktSize = strlen(lpPkt) + 1;
    
    } else {
      memcpy (lpPkt, &gMyUDPPortN, sizeof(gMyUDPPortN));
      snprintf(lpPkt+2, sizeof(lpPkt), "Hello EE122 %d", lPktCount + 1);
      lPktSize = 2 + strlen(lpPkt) + 1;
    }
    
    printf ("Sending packet (%d bytes) %d\n", lPktSize, lPktCount+1);
    lNumBytesSent = MNL_sendto(gSkt, lpPkt, lPktSize, 0/*flags*/, (struct sockaddr *) &gDestSockAddr, sizeof(gDestSockAddr), 0, 0, 0, &lErrFlag);
    if (lErrFlag == 1) {
      printf ("Error: Internal error in MNL_sendto : %s\n", strerror(errno));
    }
    // Check lNummByteSent as you normally check the return value of sendto

    if (iWaitForResponse == 1) {
      WaitForReply();
    }
  }
}

/**
  * Prints the usage information.
  */
void usage (const char* ipProgName) {
    printf (
                "\nUsage: %s <listen port>\n"
                "\n"
                ,
                ipProgName
            );
}

int main(int argc, char* argv[]) {

  if (argc < 2) {
    usage(argv[0]);
    exit(1);
  }

  gDestPortH = atoi (argv[1]);
  if (gDestPortH == 0) {
    printf("Invalid port number : %s\n", argv[1]);
    exit (1);
  }
 
  if (MNLClientInit() != 1) {
    printf ("Unable to initialize My Network Layer. Abort.\n");
    exit (1);
  }

  InitSkt();

  //SendPkts(MAX_PKT_LEN, 10, 0); 
  //SendPkts(0, 50, 0);
  SendPkts(0, 50, 1);
  //SendPkts(0, 1000, 0);

}

